/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * some interface functions
 * the main idea of the interface is the possibility to use it keyboard only.
 *
 * note : i don't know how to handle X-window errors, if someone could help...
 */

#include "interface.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xm.h"
#include "mixer.h"
#include "screen.h"
#include "sub_interface.h"
#include "keyboard_definition.h"
#include "live_player.h"
#include "file_selector.h"
#include "instrument.h"
#include "sample.h"

void init_menu_pixmap(INTERFACE *i)
{
  void menu_around(INTERFACE *, int, int);
  draw_box(i->X_ecran, i->menu_pixmap, i->context, 2, 2, 296, 50, 8);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 8, 8+38, "Xsoundtrack", i->title, 9);//yellow
  draw_box(i->X_ecran, i->menu_pixmap, i->context, 2+10, 8+50, 296-20, 50+18, 9);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 24, 75,
	      "Hope you'll enjoy using this", i->medium, 10);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 24, 95,
	      "program as much as i enjoyed", i->medium, 10);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 24, 115,
	      "making it.              Sed.", i->medium, 10);

  draw_string(i->X_ecran, i->menu_pixmap, i->context, 50, 175,
	      "This program is total free software.", i->little, 10);//red

  draw_3d_box(i->X_ecran, i->menu_pixmap, i->context, 2+10, 55+80, 60, 25, 3, UP_3D);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 8+10, 75+80, "New", i->big, 0);

  draw_3d_box(i->X_ecran, i->menu_pixmap, i->context, 2+25+80, 55+80, 70, 25, 3, UP_3D);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 8+25+80, 75+80, "Load", i->big, 0);

  draw_3d_box(i->X_ecran, i->menu_pixmap, i->context, 2+40+80+90, 55+80, 70, 25, 3, UP_3D);
  draw_string(i->X_ecran, i->menu_pixmap, i->context, 8+40+80+90, 75+80, "Exit", i->big, 0);

  menu_around(i, 0, UP_3D);
}

/* create the menu-window */
/* return 0 on error 1 otherwise */
int init_interface(INTERFACE *interface)
{
  XGCValues val;
  Display *X_ecran;

  /* don't forget the needed memset */
  memset(interface, 0, sizeof(INTERFACE));

  /* open the screen */
  if (!(X_ecran=XOpenDisplay(0)))
    return 0;

  /* init the colors */
  init_colors(X_ecran);

  /* create the graphic context (for drawing) */
  val.foreground=WhitePixel(X_ecran, DefaultScreen(X_ecran));
  val.background=BlackPixel(X_ecran, DefaultScreen(X_ecran));
  interface->context=XCreateGC(X_ecran, DefaultRootWindow(X_ecran), GCBackground|GCForeground, &val);

  /* load the needed fonts */
  interface->little=XLoadFont(X_ecran, "-adobe-courier-medium-r-normal--10-*-*-*-*-*-iso8859-1");
  interface->medium=XLoadFont(X_ecran, "-adobe-courier-medium-r-normal--15-*-*-*-*-*-iso8859-1");
  interface->big=XLoadFont(X_ecran, "-adobe-courier-medium-r-normal--25-*-*-*-*-*-iso8859-1");
  interface->title=XLoadFont(X_ecran, "-adobe-utopia-medium-r-normal--50-*-*-*-*-*-iso8859-1");

  /* create the menu window */
  interface->menu=XCreateSimpleWindow(X_ecran, DefaultRootWindow(X_ecran),
				     0, 0, 300, 180, 3, 3,
				     BlackPixel(X_ecran, DefaultScreen(X_ecran)));
  XMapWindow(X_ecran, interface->menu);
  XSelectInput(X_ecran, interface->menu, KeyPressMask|ExposureMask);
  {
    /* put the name of the window (i am not sure of this code, shoud read books to know if it's correct */
    XTextProperty proper;
    char *z="Xsoundtrack : menu";
    if (!XStringListToTextProperty(&z, 1, &proper))
      fprintf(stderr, "interface.init_interface error with XStringToTextProperty\n");
    else {
      XSetWMName(X_ecran, interface->menu, &proper);
      XFree(proper.value);
    }
  }

  /* the pixmap */
  interface->menu_pixmap=XCreatePixmap(X_ecran, DefaultRootWindow(X_ecran), 300, 180,
				       DefaultDepth(X_ecran, DefaultScreen(X_ecran)));

  interface->X_ecran=X_ecran;

  interface->xms=(XM **)0;
  interface->nb_xms=0;
  interface->is_playing=0;
  interface->current_menu_button=0;

  init_menu_pixmap(interface);

  return 1;
}

void close_interface(INTERFACE *i)
{
  XCloseDisplay(i->X_ecran);
}

void new_xm_menu(INTERFACE *i)
{
  XM *new_xm=(XM *)malloc(sizeof(XM));
  XM **buf=i->xms;

  if (!new_xm) {
    fprintf(stderr, "interface.new_xm_menu : warning malloc failed 1\n");
    return;
  }

  /* ok, we gonna put some default-data : 1 pattern, song-length=1, 0 instrument, default bpm=125, default speed=6, so on */
  memset(new_xm, 0, sizeof(XM));

  new_xm->module_name[0]=0;
  strcpy(new_xm->tracker_name, "Xsoundtrack 1.0     ");
  new_xm->song_length=1;
  new_xm->restart=0;
  new_xm->nb_channels=6;
  new_xm->nb_patterns=1;
  new_xm->nb_instruments=0;
  new_xm->default_tempo=6;
  new_xm->default_bpm=125;
  new_xm->patterns_order[0]=0;
  new_xm->patterns_nb_rows[0]=64;
  new_xm->patterns[0]=(note *)malloc(64*6*sizeof(note));
  if (!new_xm->patterns[0]) {
    free(new_xm);
    fprintf(stderr, "interface.new_xm_menu : warning malloc failed 2\n");
    return;
  }
  memset(new_xm->patterns[0], 0xff, 64*6*sizeof(note));
  {
    int i;
    for (i=0; i<64*6; i++)
      new_xm->patterns[0][i].volume=0,
      new_xm->patterns[0][i].effect_parameter=0;
  }

  buf=(XM **)realloc(buf, (i->nb_xms+1)*sizeof(XM *));
  if (!buf) {
    free(new_xm->patterns[0]);
    free(new_xm);
    fprintf(stderr, "interface.new_xm_menu : warning realloc failed\n");
    return;
  }

  buf[i->nb_xms++]=new_xm;

  init_xm_data(new_xm);

  if (!init_xm_interface(i, new_xm)) {
    free_xm(new_xm);
    fprintf(stderr, "interface.new_xm_menu : warning error while initing the interface\n");
    return;
  }

  i->xms=buf;

  put_name_xm(i, new_xm);
}

void load_xm_menu(INTERFACE *i)
{
  XM **buf=i->xms;
  XM *xm=(XM *)malloc(sizeof(XM));

  if (!xm) {
    fprintf(stderr, "interface.load_xm_menu warning : malloc failed\n");
    return;
  }

  memset(xm, 0, sizeof(XM));

  if (!init_file_selector(i, ".")) {
    fprintf(stderr, "interface.load_xm.menu error : could not deal with the file selector\n");
    return;
  }

  if (!i->file_selector.out_name)
    return;

  if (!load_xm(i->file_selector.out_name, xm)) {
    fprintf(stderr, "interface.load_xm_menu warning : unable to load the specified file, sorry\n");
    return;
  }

  buf=(XM **)realloc(buf, (i->nb_xms+1)*sizeof(XM *));
  if (!buf) {
    free_xm(xm);
    fprintf(stderr, "interface.load_xm_menu : warning realloc failed\n");
    return;
  }

  buf[i->nb_xms++]=xm;

  init_xm_data(xm);
  if (!init_xm_interface(i, xm)) {
    free_xm(xm);
    fprintf(stderr, "interface.load_xm_menu : warning error while initing the interface\n");
    return;
  }

  i->xms=buf;

  xm->file_name=strdup(i->file_selector.out_name);

  put_name_xm(i, xm);
}

void exit_menu(INTERFACE *i)
{
  /* will have to change it later... */
  exit(0);
}

/* draw the current around button in the menu window */
void menu_around(INTERFACE *i, int button, int type)
{
  int x, y, width, height;

  switch(button) {
    default : /* no default but gcc is not happy */
    case 0 :
      x=2+10-3;
      y=55+80-3;
      width=60+6;
      height=25+6;
      break;
    case 1 :
      x=2+25+80-3;
      y=55+80-3;
      width=70+6;
      height=25+6;
      break;
    case 2 :
      x=2+40+80+90-3;
      y=55+80-3;
      width=70+6;
      height=25+6;
      break;
  }

  draw_rectangle(i->X_ecran, i->menu_pixmap, i->context, x, y, width, height, 2, type==IN ? 1 : 0);
  draw_rectangle(i->X_ecran, i->menu, i->context, x, y, width, height, 2, type==IN ? 1 : 0);
}

/* draw the menu-window */
/* look this bad code ! should make generic functions, nothing should be hard-coded like this... */
void draw_menu(INTERFACE *i)
{
  XCopyArea(i->X_ecran, i->menu_pixmap, i->menu, i->context, 0, 0, 300, 180, 0, 0);
}

/* redraw one button of the menu. Used when you push a button. */
void redraw_button_menu(INTERFACE *i, int button, int type)
{
  switch(button) {
    case 0 :
      draw_3d_box_around(i->X_ecran, i->menu_pixmap, i->context, 2+10, 55+80, 60, 25, 3, type);
      draw_3d_box_around(i->X_ecran, i->menu, i->context, 2+10, 55+80, 60, 25, 3, type);
      break;
    case 1 :
      draw_3d_box_around(i->X_ecran, i->menu_pixmap, i->context, 2+25+80, 55+80, 70, 25, 3, type);
      draw_3d_box_around(i->X_ecran, i->menu, i->context, 2+25+80, 55+80, 70, 25, 3, type);
      break;
    case 2 :
      draw_3d_box_around(i->X_ecran, i->menu_pixmap, i->context, 2+40+80+90, 55+80, 70, 25, 3, type);
      draw_3d_box_around(i->X_ecran, i->menu, i->context, 2+40+80+90, 55+80, 70, 25, 3, type);
      break;
  }
  XFlush(i->X_ecran);
}

void keypress_menu(INTERFACE *i, XKeyEvent event)
{
  KeySym ks=XLookupKeysym(&event, 0);

  if (ks==NoSymbol) {
    fprintf(stderr, "interface.keypress_menu warning ks==NoSymbol\n");
    return;
  }

  switch (ks) {
    case XK_Left :
      menu_around(i, i->current_menu_button, OUT);
      i->current_menu_button--;
      if (i->current_menu_button<0)
	i->current_menu_button=2;
      menu_around(i, i->current_menu_button, IN);
      break;
    case XK_Right :
      menu_around(i, i->current_menu_button, OUT);
      i->current_menu_button++;
      if (i->current_menu_button>2)
	i->current_menu_button=0;
      menu_around(i, i->current_menu_button, IN);
      break;
    case XK_Return :
    case XK_space :
      redraw_button_menu(i, i->current_menu_button, DOWN_3D);
      switch (i->current_menu_button) {
	case 0 :
	  menu_around(i, 0, OUT);
	  new_xm_menu(i);
	  menu_around(i, 0, IN);
	  break;
	case 1 :
	  menu_around(i, 1, OUT);
	  load_xm_menu(i);
	  menu_around(i, 1, IN);
	  break;
	case 2 :
	  menu_around(i, 2, OUT);
	  exit_menu(i);
	  menu_around(i, 2, IN);
	  break;
      }
      redraw_button_menu(i, i->current_menu_button, UP_3D);
      break;
  }
}

/* handling of events */
/* in the menu window */
void handle_menu(INTERFACE *i, XEvent event)
{
  switch (event.type) {
    case Expose :
      if (event.xexpose.count==0)
	draw_menu(i);
      break;
    case KeyPress :
      if (i->file_selector.in_use)
	break;
      keypress_menu(i, event.xkey);
      break;
  }
}

void draw_restart(INTERFACE *i, XM *xm)
{
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, ADX+5, ADY+5, 148-7*6-4, 74-1, 6, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9, ADY+1+5+4+8+11-5, "Number of", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9+6, ADY+1+5+4+8+11+5, "channels", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9+6*2, ADY+1+5+4+13+3+13+6+8+11-5-1, "Restart", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9+6, ADY+1+5+4+13+3+13+6+8+11+5-1, "position", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5+13+3, ADY+1+5+4+8, 6*6-4-16, 12, 3);
  draw_box(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5+13+3, ADY+1+5+4+13+3+13+6+8-1, 6*6-4-16, 12, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5, ADY+1+5+4, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5, ADY+1+5+4+13+3, UP_3D, ARROW_DOWN);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5, ADY+1+5+4+13+3+13+6-1, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5, ADY+1+5+4+13+3+13+6+13+3-1, UP_3D, ARROW_DOWN);
}

void init_xm_pixmap(INTERFACE *i, XM *xm)
{
  /* the pattern-editing zone */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 150+2*16, 590, 245-2*16, 6, UP_3D);
  draw_box(i->X_ecran, xm->pixmap, i->context, 9, 144+7*18-14+2+2*9, 590-8, 16-4, 3);

  /* the instrument/sample zone */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 286+138+4, 11, 165, 101, 6, UP_3D);
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 286+138+4, 120+5, 165, 53, 6, UP_3D);
  /* instruments arrows */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 400-6, 5, 30, 1+140+2*16, 6, UP_3D);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 402-3, 11+1, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 402-3, 97+1, UP_3D, ARROW_DOWN);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 400-6+30-18, 44+4, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 400-6+30-18, 68-4, UP_3D, ARROW_DOWN);
  /* samples arrows */
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 400-6+30-18, 39+90-5-1+13, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 400-6+30-18, 61+90-5+1+5, UP_3D, ARROW_DOWN);
  /* strings for instruments/samples */
  draw_string(i->X_ecran, xm->pixmap, i->context, 470, 7, "Instruments", i->little, 1);
  draw_string(i->X_ecran, xm->pixmap, i->context, 435, 121, "Samples of current instr.", i->little, 1);

  /* Song name place */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 250-2, 5, 142, 32, 6, UP_3D);
  draw_box(i->X_ecran, xm->pixmap, i->context, 262-5-14+12+6, 19, 165-24-10-6, 14, 3);
  draw_string(i->X_ecran, xm->pixmap, i->context, 255-2-14+12, 15, "Song name :", i->little, 0);

  /* pattern / song length place */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, NNX+248, 41, 142-24, 137, 6, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252, 52, "Patterns", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, NNX+252, 52+2, 7*6, 11*8, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+7*6+4+6, 52+4, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+7*6+4+6, 51+11*8-13, UP_3D, ARROW_DOWN);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+7*6+4, 52+4+11*2, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+7*6+4, 52+4+11*2+13+3, UP_3D, ARROW_DOWN);
  draw_box(i->X_ecran, xm->pixmap, i->context, NNX+254, 52+2+3*11, 7*6-4, 11, 6);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252, 137+41-18, "Song length", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-4+13+5, 137+41-25+16-12-10-3+8, 7*6-4-16, 12, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-2, 137+41-25+16-12-10-3, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-2, 137+41-25+16-12-10+13, UP_3D, ARROW_DOWN);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252+11*6, 137+41-18-8*10, "Current", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252+11*6, 137+41-18-7*10, "pattern", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252+12*6, 137+41-18-6*10, "length", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-4+13+5, 137+41-25+16-25+8-40, 7*6-4-16, 12, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-2, 137+41-25+16-12-10-3-40, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, NNX+252+35-4+7*6-2, 137+41-25+16-12-10+13-40, UP_3D, ARROW_DOWN);

  /* nb lines shown / nb columns shown place */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 5, 148-6*6, 74, 6, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9+6*2, 1+5+4+8+11-5, "Nb lines", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9+6*5, 1+5+4+8+11+5, "shown", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9, 1+5+4+13+3+13+6+8+11-5, "Nb columns", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9+6*5, 1+5+4+13+3+13+6+8+11+5, "shown", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, 9+6*10+5+13+3, 1+5+4+8, 7*6-4-16, 12, 3);
  draw_box(i->X_ecran, xm->pixmap, i->context, 9+6*10+5+13+3, 1+5+4+13+3+13+6+8, 7*6-4-16, 12, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 9+6*10+5, 1+5+4, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 9+6*10+5, 1+5+4+13+3, UP_3D, ARROW_DOWN);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 9+6*10+5, 1+5+4+13+3+13+6, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, 9+6*10+5, 1+5+4+13+3+13+6+13+3, UP_3D, ARROW_DOWN);

  /* default bpm / default speed */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, XDD+5, 5, 148-6*9, 74, 6, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9, 1+5+4+8+11-5, "Default", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9+6*4, 1+5+4+8+11+5, "BPM", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9, 1+5+4+13+3+13+6+8+11-5, "Default", i->little, 0);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9+6*2, 1+5+4+13+3+13+6+8+11+5, "speed", i->little, 0);
  draw_box(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5+13+3, 1+5+4+8, 7*6-4-16, 12, 3);
  draw_box(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5+13+3, 1+5+4+13+3+13+6+8, 7*6-4-16, 12, 3);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5, 1+5+4, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5, 1+5+4+13+3, UP_3D, ARROW_DOWN);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5, 1+5+4+13+3+13+6, UP_3D, ARROW_UP);
  draw_arrow(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5, 1+5+4+13+3+13+6+13+3, UP_3D, ARROW_DOWN);

  /* frequence table */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 79+4, 35*6+8, 18, 6, UP_3D);
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5+6*18-5, 79+6+2, 10, 10, 3, UP_3D);
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5+6*29-5, 79+6+2, 10, 10, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9, 79+4+12, "Frequence table    Linear     Amiga", i->little, 0);

  /* the buttons */
  /* sample editor */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 79+4+18+4, 6*25+4+3, 18, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 8+2*6, 79+4+18+4+12, "Current sample editor", i->little, 0);
  /* instrument editor */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 79+26+18+4, 6*25+4+3, 18, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 8, 79+26+18+4+12, "Current instrument editor", i->little, 0);
  /* effects box */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5, 79+26+18+4+18+4, 6*12+4+3, 18, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 8, 79+26+18+4+18+4+12, "Effects menu", i->little, 0);
  /* save */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5+12*6+4+7, 79+26+18+4+18+4, 6*4+4+3, 18, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 8+12*6+8+3, 79+4+18+4+12+18+4+18+4, "Save", i->little, 0);
  /* close */
  draw_3d_box(i->X_ecran, xm->pixmap, i->context, 5+12*6+4+7+4*6+4+7, 79+26+18+4+18+4, 6*5+4+5, 18, 3, UP_3D);
  draw_string(i->X_ecran, xm->pixmap, i->context, 8+12*6+4+7+3+4*6+9, 79+4+18+4+12+18+4+18+4, "Close", i->little, 0);

  /* restart pos / nb channels */
  draw_restart(i, xm);
}

static int axm[15][4]={
  {5-2, 150+2*16-2, 590+4, 245-2*16+4},				/*  0 */
  {5-2, 5-2, 148-6*6+4, 74+4},					/*  1 */
  {XDD+5-2, 5-2, 148-6*9+4, 74+4},				/*  2 */
  {250-2-2, 5-2, 142+4, 32+4},					/*  3 */
  {400-6-2, 5-2, 30+4, 1+140+2*16+4},				/*  4 */
  {286+138+4-2, 11-2, 165+3, 101+4},				/*  5 */
  {286+138+4-2, 120+5-2, 165+3, 53+4},				/*  6 */
  {5-2, 79+4-2, 35*6+8+4, 18+4},				/*  7 */
  {5-2, 79+4+18+4-2, 6*25+4+3+4, 18+4},				/*  8 */
  {5+12*6+4+7-2, 79+26+18+4+18+4-2, 6*4+4+3+4, 18+4},		/*  9 */
  {5-2, 79+26+18+4-2, 6*25+4+3+4, 18+4},			/* 10 */
  {5+12*6+4+7+4*6+4+7-2, 79+26+18+4+18+4-2, 6*5+4+5+4, 18+4},	/* 11 */
  {5-2, 79+26+18+4+18+4-2, 6*12+4+3+4, 18+4},			/* 12 */
  {NNX+248-2, 41-2, 142-24+4, 137+4},				/* 13 */
  {ADX+5-2, ADY+5-2, 148-7*6-4+4, 74-1+4},			/* 14 */
};

unsigned long col_xm[15]= {
  6,
  6,
  6,
  6,
  6,
  6,
  6,
  6,
  3,
  3,
  3,
  3,
  3,
  6,
  6,
};

/* push or pop a subwindow (like a button) */
void up_or_down_xm(INTERFACE *i, XM *xm, int subwin, int type)
{
  draw_3d_box_around(i->X_ecran, xm->window, i->context,
		     axm[subwin][0]+2, axm[subwin][1]+2, axm[subwin][2]-4, axm[subwin][3]-4, col_xm[subwin], type);
  draw_3d_box_around(i->X_ecran, xm->pixmap, i->context,
		     axm[subwin][0]+2, axm[subwin][1]+2, axm[subwin][2]-4, axm[subwin][3]-4, col_xm[subwin], type);
}

/* draw a rectangle around the current sub-window */
/* type = IN or OUT */
void around_xm(INTERFACE *i, XM *xm, int subwin, int type)
{

  if (subwin==5||subwin==6) {			/* instruments and samples windows are special */
    draw_rectangle(i->X_ecran, xm->window, i->context, axm[subwin][0], axm[subwin][1],
		   axm[subwin][2]-1, axm[subwin][3]-1, 1, type==IN ? 1 : 0);
    draw_rectangle(i->X_ecran, xm->pixmap, i->context, axm[subwin][0], axm[subwin][1],
		   axm[subwin][2]-1, axm[subwin][3]-1, 1, type==IN ? 1 : 0);
    return;
  }

  draw_rectangle(i->X_ecran, xm->window, i->context, axm[subwin][0], axm[subwin][1],
		 axm[subwin][2], axm[subwin][3], 2, type==IN ? 1 : 0);
  draw_rectangle(i->X_ecran, xm->pixmap, i->context, axm[subwin][0], axm[subwin][1],
		 axm[subwin][2], axm[subwin][3], 2, type==IN ? 1 : 0);
  XFlush(i->X_ecran);
}

char *name_note[12]={
  "C ", "Db", "D ", "Eb", "E ", "F ", "Fb", "G ", "Ab", "A ", "Bb", "B "
};

void init_xm_dynamic(INTERFACE *i, XM *xm)
{
  int in;
  char res[25];
  void draw_patterns_(INTERFACE *, XM *);

  /* draw the instruments/samples names */
  {
    instrument *inst;

    for (in=0; in<8; in++) {
      inst=xm->instruments[xm->start_5+in];

      if (in==xm->cur_pos_5)
	draw_box(i->X_ecran, xm->pixmap, i->context, 138+294-1, 22+in*12-8, 6*26+4-1, 10, 3);
      sprintf(res, "%2.2X", (unsigned char)(in+xm->start_5+1));
      draw_string(i->X_ecran, xm->pixmap, i->context, 138+294, 22+in*12, res, i->little, 8);
      if (inst)
	draw_string(i->X_ecran, xm->pixmap, i->context, 138+294+6*4, 22+in*12, inst->name, i->little, 8);
    }

    inst=xm->instruments[xm->start_5+xm->cur_pos_5];
    for (in=0; in<4; in++) {
      if (in==xm->cur_pos_6)
	draw_box(i->X_ecran, xm->pixmap, i->context, 138+294-1, 10+20+8*12+10+in*12-8, 6*26+4-1, 10, 3);
      sprintf(res, "%2.2X", (unsigned char)(in+xm->start_6+1));
      draw_string(i->X_ecran, xm->pixmap, i->context, 138+294, 10+20+8*12+10+in*12, res, i->little, 8);
      if (inst)
	draw_string(i->X_ecran, xm->pixmap, i->context, 138+294+4*6, 10+20+8*12+10+in*12,
		    inst->sample_headers[in].sample_name, i->little, 8);
    }
  }

  /* draw the pattern order */
  for (in=0; in<8; in++) {
    if (in+xm->current_pattern-3>=xm->song_length)
      break;
    if (in+xm->current_pattern-3>=0) {
      sprintf(res, "%2.2X %2.2X", (unsigned char)(in+xm->current_pattern-3),
	      (unsigned char)(xm->patterns_order[in+xm->current_pattern-3]));
      draw_string(i->X_ecran, xm->pixmap, i->context, NNX+255+3, 52+11+in*11, res, i->little, 8);
    }
  }
  sprintf(res, "%d", xm->patterns_nb_rows[(int)xm->patterns_order[0]]);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252+31+38+20+6*(3-strlen(res)), 137+25-40, res, i->little, 8);
  sprintf(res, "%d", xm->song_length);
  draw_string(i->X_ecran, xm->pixmap, i->context, NNX+252+31+7*6-4+20+6*(3-strlen(res)), 137+25, res, i->little, 8);

  /* draw the patterns */
  draw_patterns_(i, xm);

  /* draw the song name */
  draw_string(i->X_ecran, xm->pixmap, i->context, 265-5-14+12+6, 29, xm->module_name, i->little, 8);

  /* the nb of lines/columns currently drawn for the pattern editor */
  sprintf(res, "%d", xm->nb_lines_shown);
  draw_string(i->X_ecran, xm->pixmap, i->context, 9+6*10+5+13+3+2+6*(3-strlen(res)), 1+5+4+8+10, res, i->little, 8);
  sprintf(res, "%d", xm->nb_columns_shown);
  draw_string(i->X_ecran, xm->pixmap, i->context, 69+23+6*(3-strlen(res)), 1+5+4+13+3+13+6+8+10, res, i->little, 8);

  /* default bpm/speed */
  sprintf(res, "%d", xm->default_bpm);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5+13+3+2+6*(3-strlen(res)), 1+5+4+8+10, res, i->little, 8);
  sprintf(res, "%d", xm->default_tempo);
  draw_string(i->X_ecran, xm->pixmap, i->context, XDD+9+6*7+5+13+3+2+6*(3-strlen(res)), 39+24, res, i->little, 8);

  /* nb of channels / restart position */
  sprintf(res, "%d", xm->nb_channels);
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5+13+3+2+6*(2-strlen(res)), ADY+1+27, res, i->little, 8);
  sprintf(res, "%2.2X", (unsigned char)(xm->restart));
  draw_string(i->X_ecran, xm->pixmap, i->context, ADX+9+6*9+5+13+3+2, ADY+10+26+3+6+17, res, i->little, 8);

  /* linear / amiga table */
  if (xm->flags & 1) /* linear */
    draw_3d_box_around(i->X_ecran, xm->pixmap, i->context, 5+6*18-5, 79+6+2, 10, 10, 3, DOWN_3D);
  else /* amiga */
    draw_3d_box_around(i->X_ecran, xm->pixmap, i->context, 5+6*29-5, 79+6+2, 10, 10, 3, DOWN_3D);
  
  around_xm(i, xm, xm->subwindow, IN);
}

/* put the name of the window (i am not sure of this code, shoud read books to know if it's correct */
void put_name_xm(INTERFACE *i, XM *xm)
{
  /* should handle errors, but don't know yet how to do it (after something like 2 years X-coding... do you believe it ?) */
  XTextProperty proper;
  char z[200];
  char *d=&z[0];
  int xn;
  for (xn=0; ;xn++)
    if (xm==i->xms[xn])
      break;
  sprintf(z, "Xsoundtrack : module editor - Xm %2.2x", xn+1);
  if (!XStringListToTextProperty(&d, 1, &proper))
    fprintf(stderr, "interface.put_name_xm error with XStringToTextProperty\n");
  else {
    XSetWMName(i->X_ecran, xm->window, &proper);
    XFree(proper.value);
  }
}

/* create a xm editing window, register the events... */
int init_xm_interface(INTERFACE *i, XM *xm)
{
  Display *X_ecran=i->X_ecran;

  /* sanity check */
  if (!X_ecran)
    return 0;

  /* crate the window */
  xm->window=XCreateSimpleWindow(X_ecran, DefaultRootWindow(X_ecran),
				 /* bad those hard-coded values */
				 0, 0, 600, 400, 3, 3,
				 BlackPixel(X_ecran, DefaultScreen(X_ecran)));

  /* show it */
  XMapWindow(X_ecran, xm->window);

  xm->pixmap=XCreatePixmap(X_ecran, DefaultRootWindow(X_ecran), 600, 400,
				       DefaultDepth(X_ecran, DefaultScreen(X_ecran)));

  /* select inputs */
  /* we need exposure, keyboard push, keyboard release, for now (should think about mouse and entering/leaving a window) */
  XSelectInput(X_ecran, xm->window, KeyPressMask|KeyReleaseMask|ExposureMask|FocusChangeMask);

  init_xm_pixmap(i, xm);

  xm->mode=MOVING;
  xm->subwindow=0;		/* we make a memset(0), so not really usefull, but i like do sure things */
  xm->nb_lines_shown=16;
  xm->nb_columns_shown=6;
  xm->start_column_0=0;
  xm->cursor_column_0=0;
  xm->cur_pos_0=0;
  xm->up_subwin_0=13;
  xm->cur_pos_1=0;
  xm->cur_pos_2=0;
  xm->cur_pos_4=0;
  xm->left_subwin_4=3;
  xm->right_subwin_4=5;
  xm->start_5=0;
  xm->cur_pos_5=0;
  xm->start_6=0;
  xm->cur_pos_6=0;
  xm->cur_pos_7=0;
  xm->up_subwin_7=2;
  xm->down_subwin_7=8;
  xm->down_subwin_10=9;
  xm->cur_pos_13=0;
  xm->left_subwin_13=14;
  xm->cur_pos_14=0;
  xm->left_subwin_14=8;
  xm->new_line=0;
  xm->new_pattern=0;
  xm->current_octave=3;
  xm->cut_buffer_nb_lines=0;
  xm->cut_buffer_nb_columns=0;
  xm->current_selection_pattern=-1;
  xm->current_selection_nb_lines=0;
  xm->current_selection_start_line=0;
  xm->current_selection_nb_columns=0;
  xm->current_selection_start_column=0;
  memset(xm->live_notes, -1, 20*sizeof(note));
  init_xm_dynamic(i, xm);

  /* done */
  return 1;
}

void close_xm_interface(INTERFACE *i, XM *xm)
{
  if (!i->X_ecran)
    return;

  XDestroyWindow(i->X_ecran, xm->window);
}

/* draw an xm-window (fully) */
void draw_xm(INTERFACE *i, XM *xm)
{
  XCopyArea(i->X_ecran, xm->pixmap, xm->window, i->context, 0, 0, 600, 400, 0, 0);
}

#define is_in_playmode(xm) (xm->subwindow==0 && xm->mode==MOVING)

/* a global var for sub_interface.interface_subwin_0 which wants to know shift, control state */
unsigned int status;

void keypress_xm(INTERFACE *i, XM *xm, XKeyEvent event)
{
  char dum[50];
  KeySym ks;

  status=event.state;

  XLookupString(&event, dum, 50, &ks, (XComposeStatus *)0);

  if (ks==NoSymbol) {
    fprintf(stderr, "interface.keypress_xm : warning ks=NoSymbol\n");
    return;
  }

  /* Escape will put in moving mode, any place we are */
  if (ks==XK_Escape) {
    if (xm->mode==MOVING) {
      /* let's stop the playing. */
      i->is_playing=0;
      if (i->current_xm)
	i->current_xm->mix&=~MODULE_PLAY;
      return;
    }
    xm->mode=MOVING;
    exit_subwin[xm->subwindow](i, xm);
    up_or_down_xm(i, xm, xm->subwindow, UP_3D);
    if (is_in_playmode(xm))
      XAutoRepeatOff(i->X_ecran);
    return;
  }

  if (xm->subwindow==0) {
    if (!status)
      if (ks>=XK_F1 && ks<=XK_F7) {
	xm->current_octave=ks-XK_F1;
	fprintf(stderr, "octave=%d\n", xm->current_octave);
	return;
      }
    if (xm->mode==MOVING) {
      void interface_subwin_5(INTERFACE *, XM *, KeySym);
      if (ks==XK_Page_Up) {
      interface_subwin_5(i, xm, XK_Up);
      return;
      }
      if (ks==XK_Page_Down) {
	interface_subwin_5(i, xm, XK_Down);
	return;
      }
    }
  }

  /* any place in move mode, space plays/stop the current module. If subwindow is pattern editor, do it too
   * if mode!=MOVING
   */
  if (ks==XK_space && (xm->subwindow==0 || xm->mode==MOVING)) {
    /* go in/leave the playing of the module */
    /* if playing another module, first stop it. */
    if (i->is_playing && i->current_xm!=xm)
      i->is_playing=0;
    /* now, toggle */
    i->is_playing=1-i->is_playing;
    if (i->is_playing)
      xm->mix|=MODULE_PLAY;
    else
      xm->mix&=~MODULE_PLAY;
    i->current_xm=xm;
    return;
  }

  /* if in editing mode, then call the proper function */
  if (xm->mode==EDIT) {
    interface_subwin[xm->subwindow](i, xm, ks);
    return;
  }

  /* we're in moving mode. Handle Enter and left-right-up-down */
  switch(ks) {
    case XK_Return :
      if (is_in_playmode(xm)) {
	XKeyboardControl kb;
	kb.auto_repeat_mode=AutoRepeatModeDefault;
	XChangeKeyboardControl(i->X_ecran, KBAutoRepeatMode, &kb);
      }
      xm->mode=EDIT;
      up_or_down_xm(i, xm, xm->subwindow, DOWN_3D);
      enter_subwin[xm->subwindow](i, xm);
      break;
    case XK_Left :
      switch (xm->subwindow) {			/* d'you like big switches as i do ? */
	case 0 :
	  /* nothing */
	  break;
	case 1 :
	  /* nothing */
	  break;
	case 2 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=1;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 3 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=2;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 4 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->left_subwin_4;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 5 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=4;
	  xm->right_subwin_4=5;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 6 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=4;
	  xm->right_subwin_4=6;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 7 :
	  /* nothing */
	  break;
	case 8 :
	  /* nothing */
	  break;
	case 9 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=12;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 10 :
	  /* nothing */
	  break;
	case 11 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=9;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 12 :
	  /* nothing */
	  break;
	case 13 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->left_subwin_13;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 14 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->left_subwin_14;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
      }
      break;
    case XK_Right :
      switch (xm->subwindow) {
	case 0 :
	  /* nothing */
	  break;
	case 1 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=2;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 2 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=3;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 3 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=4;
	  xm->left_subwin_4=3;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 4 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->right_subwin_4;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 5 :
	  /* nothing */
	  break;
	case 6 :
	  /* nothing */
	  break;
	case 7 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=13;
	  xm->left_subwin_13=7;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 8 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=14;
	  xm->left_subwin_14=8;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 9 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=11;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 10 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=14;
	  xm->left_subwin_14=10;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 11 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=14;
	  xm->left_subwin_14=11;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 12 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=9;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 13 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=4;
	  xm->left_subwin_4=13;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 14 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=13;
	  xm->left_subwin_13=14;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
      }
      break;
    case XK_Up :
      switch (xm->subwindow) {
	case 0 : {
	  XKeyboardControl kb;
	  kb.auto_repeat_mode=AutoRepeatModeDefault;
	  XChangeKeyboardControl(i->X_ecran, KBAutoRepeatMode, &kb);
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->up_subwin_0;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	}
	case 1 :
	case 2 :
	case 3 :
	case 4 :
	case 5 :
	  /* nothing */
	  break;
	case 6 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=5;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 7 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->up_subwin_7;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 8 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=7;
	  xm->down_subwin_7=8;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 9 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=10;
	  xm->down_subwin_10=9;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 10 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=8;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 11 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=10;
	  xm->down_subwin_10=11;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 12 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=10;
	  xm->down_subwin_10=12;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 13 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=3;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 14 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=7;
	  xm->down_subwin_7=14;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
      }
      break;
    case XK_Down :
      switch (xm->subwindow) {
	case 0 :
	  /* nothing */
	  break;
	case 1 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=7;
	  xm->up_subwin_7=1;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 2 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=7;
	  xm->up_subwin_7=2;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 3 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=13;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 4 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=4;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 5 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=6;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 6 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=6;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 7 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->down_subwin_7;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 8 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=10;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 9 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=9;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 10 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=xm->down_subwin_10;
	  around_xm(i, xm, xm->subwindow, IN);
	  break;
	case 11 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=11;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 12 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=12;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 13 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=13;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
	case 14 :
	  around_xm(i, xm, xm->subwindow, OUT);
	  xm->subwindow=0;
	  xm->up_subwin_0=14;
	  around_xm(i, xm, xm->subwindow, IN);
	  XAutoRepeatOff(i->X_ecran);
	  break;
      }
      break;
    default : { /* maybe we are in playing mode and the key pushed is a note, so say the mixer to play it */
      int note=-1;
      int in;
      int chan;
      if (!is_in_playmode(xm))
	break;
      /* test down keyboard */
      for (in=0; in<17; in++)
	if (keyboard_down[in]==ks) {
	  note=in;
	  break;
	}
      /* then the up keyboard */
      if (note==-1)
	for (in=0; in<20; in++)
	  if (keyboard_up[in]==ks) {
	    note=in+12;
	    break;
	  }
      if (note==-1)
	break;
      note+=xm->current_octave*12;
      if (note>=96)
	break;
      //fprintf(stderr, "ok note=%d\n", note);
      chan=get_free_live_player_channel(xm);
      if (chan==-1) {
	XBell(i->X_ecran, 0);
	break;
      }
      put_live_player_note(xm, chan, note);
      open_live_player(xm);
      /* don't forget that */
      i->current_xm=xm;
      break;
    }
  }
}

/* handle a key release. Used only when in playmode to make a cut key.
 * Note : when we leave the window, it's as if we released all the keys. That's for simplicity.
 */
void keyrelease_xm(INTERFACE *i, XM *xm, XKeyEvent event)
{
  int note=-1;
  int in;
  int chan;
  char dum[50];
  KeySym ks;

  XLookupString(&event, dum, 50, &ks, (XComposeStatus *)0);

  if (ks==NoSymbol) {
    fprintf(stderr, "interface.keyrelease_xm : warning ks=NoSymbol\n");
    return;
  }

  /* test down keyboard */
  for (in=0; in<17; in++)
    if (keyboard_down[in]==ks) {
      note=in;
      break;
    }
  /* then the up keyboard */
  if (note==-1)
    for (in=0; in<20; in++)
      if (keyboard_up[in]==ks) {
	note=in+12;
	break;
      }
  if (note==-1)
    return;
  note+=xm->current_octave*12;
  if (note>=96)
    return;
  chan=find_live_player_channel(xm, note);
  if (chan==-1)
    return;
  /* ok, well, put the cut-key */
  put_live_player_note(xm, chan, 96);
  open_live_player(xm);
  /* not to be forgottent */
  i->current_xm=xm;
}

/* in a xm window */
void handle_xm(INTERFACE *i, XM *xm, XEvent event)
{
  switch (event.type) {
    case Expose :
      draw_xm(i, xm);
      break;
    case KeyPress :
      keypress_xm(i, xm, event.xkey);
      break;
    case KeyRelease :
      if (is_in_playmode(xm))
	keyrelease_xm(i, xm, event.xkey);
      break;
    case FocusIn :
      if (is_in_playmode(xm))
	XAutoRepeatOff(i->X_ecran);
      break;
    case FocusOut :
      if (is_in_playmode(xm)) {
	XKeyboardControl kb;
	kb.auto_repeat_mode=AutoRepeatModeDefault;
	XChangeKeyboardControl(i->X_ecran, KBAutoRepeatMode, &kb);
      }
      /* in any cases, close the "live"-player */
      close_live_player(xm);
      break;
  }
}

/* in a xm window */

/* sample_subwin_6 of sample.c needs the xm of the sample, so a global because don't want to change all. */
/* other funs too */
XM *the_xm;

/* the main function */
void handle_event(INTERFACE *i)
{
  XEvent event;
  int j;

  if (i->in_record || i->is_playing || (i->current_xm && (i->current_xm->mix & LIVE_PLAY))) {
    /* not very pretty the -1L, huh ? */
    if (!XCheckMaskEvent(i->X_ecran, -1L, &event))
      return;
  } else
    XNextEvent(i->X_ecran, &event);

  /* if in record, other windows only want Expose */
  if (i->in_record && i->record_window!=event.xany.window)
    if (event.type!=Expose)
      return;

  /* if file_selector is here, other windows only want Expose */
  if (i->file_selector.in_use && i->file_selector.window!=event.xany.window)
    if (event.type!=Expose)
      return;

  /* is it the menu ? */
  if (event.xany.window==i->menu)
    return handle_menu(i, event);

  /* or an xm window */
  for (j=0; j<i->nb_xms; j++)
    if (i->xms[j]->window==event.xany.window)
      return handle_xm(i, i->xms[j], event);

  /* or perhaps an instrument window */
  for (j=0; j<i->nb_xms; j++) {
    int k;
    for (k=0; k<128; k++)
      if (i->xms[j]->instruments[k]) {
	int s;
	sample *samp=&i->xms[j]->instruments[k]->sample_headers[0];
	if (i->xms[j]->instruments[k]->is_shown &&
	    i->xms[j]->instruments[k]->window==event.xany.window) {
	  the_xm=i->xms[j];
	  return instrument_handle_event(i, i->xms[j]->instruments[k], &event);
	}
	/* or a sample of this instrument */
	for (s=0; s<16; s++, samp++) {
	  if (samp->is_shown && samp->window==event.xany.window) {
	    the_xm=i->xms[j];
	    return sample_handle_event(i, samp, &event);
	  }
	}
      }
  }

  /* maybe the file selector */
  if (i->file_selector.in_use && i->file_selector.window==event.xany.window)
    return file_selector_handle_event(i, &event);

  /* don't printf, we can come here and have no matching window (the file selector is still there
   * when we load, it cans take some time, user can push keys and so events are generated. Same
   * goes for when we close a xm).
   * And i don't think we could receive an event for a window that doesn't belong to xsoundtrack (yes, X11
   * is bugged, but bugs like that can't exist (well, hope so...)).
   */
  //fprintf(stderr, "interface.handle_event warning : no window match %ld\n", event.xany.window);
}
