/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * the screen stuff (draw things, handle events)
 *
 * note : i absolutly don't know how to handle X-window errors, so if it does not work properly with your
 * machine, let me know please. If you know how to handle errors, let me know too. I want to make clean
 * things.
 */

#include "screen.h"

#include <X11/Xlib.h>
#include <string.h>

#include "interface.h"

unsigned long colors[11];
/*
 * 0 = black
 * 1 = white
 * 2, 3, 4 = lightblue+ lightblue lightblue-
 * 5, 6, 7 = grey+ grey grey-
 * 8 = blue
 * 9 = yellow
 * 10 = red
 */

void init_colors(Display *X_ecran)
{
  unsigned long visu;
  #define CNI 3
  static unsigned long the_cols[11][3]={
    {0, 0, 0},
    {65535, 65535, 65535},
    {173*256+((173*256)/CNI), 216*256+((216*256)/CNI), 230*256+((230*256)/CNI)},
    {173*256, 216*256, 230*256},
    {173*256-((173*256)/CNI), 216*256-((216*256)/CNI), 230*256-((230*256)/CNI)},
    {190*256+((190*256)/CNI), 190*256+((190*256)/CNI), 190*256+((190*256)/CNI)},
    {190*256, 190*256, 190*256},
    {190*256-((190*256)/CNI), 190*256-((190*256)/CNI), 190*256-((190*256)/CNI)},
    {0, 0, 65535},
    {65535, 65535, 0},
    {65535, 0, 0},
  };

  /* hope this mechanism works in all cases. Hope it does not look too bad on greyscales screens. */
  if ((visu=DefaultVisual(X_ecran, DefaultScreen(X_ecran))->class)==StaticGray || visu==StaticColor ||
      visu==TrueColor) {
    /* here, we have statical things. We must query colors. */
    int i;
    XColor col;
    for (i=0; i<11; i++) {
      col.red=the_cols[i][0]>65535 ? 65535 : the_cols[i][0];
      col.green=the_cols[i][1]>65535 ? 65535 : the_cols[i][1];
      col.blue=the_cols[i][2]>65535 ? 65535 : the_cols[i][2];
      XAllocColor(X_ecran, DefaultColormap(X_ecran, DefaultScreen(X_ecran)), &col);
      colors[i]=col.pixel;
      XFreeColors(X_ecran, DefaultColormap(X_ecran, DefaultScreen(X_ecran)), &col.pixel, 1, 0);
    }
  } else {
    /* here, we can alloc. */
    XColor col[11];
    int i;
    XAllocColorCells(X_ecran, DefaultColormap(X_ecran, DefaultScreen(X_ecran)), False, (unsigned long *)0, 0,
		     colors, 11);
    for (i=0; i<11; i++) {
      col[i].pixel=colors[i];
      col[i].red=the_cols[i][0]>65535 ? 65535 : the_cols[i][0];
      col[i].green=the_cols[i][1]>65535 ? 65535 : the_cols[i][1];
      col[i].blue=the_cols[i][2]>65535 ? 65535 : the_cols[i][2];
      col[i].flags=DoRed|DoGreen|DoBlue;
    }
    XStoreColors(X_ecran, DefaultColormap(X_ecran, DefaultScreen(X_ecran)), col, 11);
  }
}

/* not really a plot, because we want a width, so... anyway this function is not too much called,
 * no problem if slow a bit */
void draw_plot (Display *X_ecran, Window w, GC context, int x, int y, int width, unsigned long color)
{
  XGCValues val;

  val.foreground=colors[color];
  XChangeGC(X_ecran, context, GCForeground, &val);

  XFillRectangle(X_ecran, w, context, x, y, width, width);
}

void draw_line (Display *X_ecran, Window w, GC context, int x, int y, int x2, int y2, unsigned long color)
{
  XGCValues val;

  val.foreground=colors[color];
  val.line_width=1;
  XChangeGC(X_ecran, context, GCForeground|GCLineWidth, &val);

  XDrawLine(X_ecran, w, context, x, y, x2, y2);
}

/* color must be valid, otherwise strange things could happen */
void draw_string (Display *X_ecran, Window w, GC context, int px, int py, char *to_draw, Font f, unsigned long color)
{
  XTextItem x;
  XGCValues val;

  /* set the color */
  val.foreground=colors[color];
  XChangeGC(X_ecran, context, GCForeground, &val);

  /* draw the string */
  x.chars=to_draw;
  x.nchars=strlen(to_draw);
  x.delta=0;
  x.font=f;
  XDrawText(X_ecran, w, context, px, py, &x, 1);
}

void draw_box(Display *d, Window w, GC g, int x, int y, int width, int height, unsigned long color)
{
  XGCValues val;

  val.foreground=colors[color];
  XChangeGC(d, g, GCForeground, &val);
  XFillRectangle(d, w, g, x, y, width, height);
}

void draw_3d_box_around(Display *d, Window w, GC g, int x, int y, int width, int height, unsigned long color, int type)
{
  int color1, color2, color3;
  XGCValues val;

  color1=colors[color];

  color2=colors[color-1];

  color3=colors[color+1];

  if (type==UP_3D)
    val.foreground=color2;
  else
    val.foreground=color3;
  val.line_width=1;
  XChangeGC(d, g, GCForeground|GCLineWidth, &val);
  XDrawLine(d, w, g, x, y, x+width, y);
  XDrawLine(d, w, g, x, y+1, x, y+height);
  XDrawLine(d, w, g, x+1, y+1, x+width-1, y+1);
  XDrawLine(d, w, g, x+1, y+2, x+1, y+height-1);

  if (type==UP_3D)
    val.foreground=color3;
  else
    val.foreground=color2;
  XChangeGC(d, g, GCForeground, &val);
  XDrawLine(d, w, g, x+width-1, y+1, x+width-1, y+height);
  XDrawLine(d, w, g, x+1, y+height-1, x+width-1, y+height-1);
  XDrawLine(d, w, g, x+width-2, y+2, x+width-2, y+height-1);
  XDrawLine(d, w, g, x+2, y+height-2, x+width-2, y+height-2);
}

/* this function draws a "3d"-box, ie a square with around it line lighter ok darker to simulate 3d */
void draw_3d_box(Display *d, Window w, GC g, int x, int y, int width, int height, unsigned long color, int type)
{
  int color1, color2, color3;
  XGCValues val;

  color1=colors[color];

  color2=colors[color-1];

  color3=colors[color+1];

  val.foreground=color1;
  XChangeGC(d, g, GCForeground, &val);
  XFillRectangle(d, w, g, x+2, y+2, width-4, height-4);

  if (type==UP_3D)
    val.foreground=color2;
  else
    val.foreground=color3;
  val.line_width=1;
  XChangeGC(d, g, GCForeground|GCLineWidth, &val);
  XDrawLine(d, w, g, x, y, x+width, y);
  XDrawLine(d, w, g, x, y+1, x, y+height);
  XDrawLine(d, w, g, x+1, y+1, x+width-1, y+1);
  XDrawLine(d, w, g, x+1, y+2, x+1, y+height-1);

  if (type==UP_3D)
    val.foreground=color3;
  else
    val.foreground=color2;
  XChangeGC(d, g, GCForeground, &val);
  XDrawLine(d, w, g, x+width-1, y+1, x+width-1, y+height);
  XDrawLine(d, w, g, x+1, y+height-1, x+width-1, y+height-1);
  XDrawLine(d, w, g, x+width-2, y+2, x+width-2, y+height-1);
  XDrawLine(d, w, g, x+2, y+height-2, x+width-2, y+height-2);
}

void draw_rectangle(Display *d, Window w, GC g, int x, int y, int width, int height, int linewidth, unsigned long color)
{
  XGCValues val;

  val.line_width=linewidth;
  val.foreground=colors[color];
  XChangeGC(d, g, GCForeground|GCLineWidth, &val);

  XDrawRectangle(d, w, g, x, y, width, height);
}

void draw_arrow(Display *d, Window w, GC g, int x, int y, int type, int mode)
{
  draw_3d_box(d, w, g, x, y, 13, 13, 3, type);

  if (mode==ARROW_UP) {
    draw_line(d, w, g, x+6, y+3,  x+4, y+9, 0);
    draw_line(d, w, g, x+4, y+9,  x+8+1, y+9, 0);
    draw_line(d, w, g, x+8, y+9,  x+6, y+3, 0);
    return;
  }
  if (mode==ARROW_DOWN) {
    draw_line(d, w, g, 12+x-6, 12+y-3,  12+x-4, 12+y-9, 0);
    draw_line(d, w, g, 12+x-4+1, 12+y-9,  12+x-8, 12+y-9, 0);
    draw_line(d, w, g, 12+x-8, 12+y-9,  12+x-6, 12+y-3, 0);
    return;
  }
  if (mode==ARROW_LEFT) {
    draw_line(d, w, g, x+3, y+6,  x+9, y+4, 0);
    draw_line(d, w, g, x+9, y+4,  x+9, y+9, 0);
    draw_line(d, w, g, x+9, y+8,  x+3, y+6, 0);
    return;
  }
  if (mode==ARROW_RIGHT) {
    draw_line(d, w, g, 12+x-9, 12+y-4,  12+x-3, 12+y-6, 0);
    draw_line(d, w, g, 12+x-9, 12+y-8,  12+x-9, 12+y-3, 0);
    draw_line(d, w, g, 12+x-9, 12+y-8,  12+x-3, 12+y-6, 0);
    /* well, without it, troubles. Surely something to configure in the GC, but... */
    draw_plot(d, w, g, 12+x-3, 12+y-6, 1, 0);
    return;
  }
}
