/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * The file contains a mixer, ie routines necessary to write the sound to the soundcard.
 */

#include "mixer.h"

#include "xm.h"
#include "card.h"
#include "live_player.h"

/* maybe not all those include are necessary, due to coding-process, which cause destruction of code after its need */
#include <stdio.h>
#include <string.h>
#include <math.h>

/* the channels */
/* have moved in each xm */

/* buffer for mixing in 32b */
int mixer_soft_buffer[1024*2/4];
int mixer_pos;
/* buffer for output in the soundcard */
char mixer_buffer_soundcard[1024];

int init_mixer(int rate, int precision, int mono_stereo)
{
  /* well, for the first player, we gonna fix the things, so only 44100/16/stereo (incremental coding) */
  //memset(channels, 0, sizeof(channels));
  return 1;
}

/* this function inits some global vars in the xm-module */
void init_xm_data(XM *module)
{
/*  int new_line(XM *); */
  module->current_pattern=0;
  module->current_line=0; /*-1;*/     /* -1 because the mixer will realise it has no more byte to handle, so increment
				* the current line, which shall reach 0
				*/
  module->current_tempo=module->default_tempo; /* why did they call it tempo ? */
  /*  module->speed=32760;*/ /* one big value to directly change the line in new_line, bad method, will have to handle
			* it differently
			*/
  /*  fprintf(stderr, "mixer.init_xm_data don't forget to handle the very first line differently or what...\n");*/

  module->bpm=module->default_bpm;
/*  module->to_handle=new_line(module); */ /* we gonna do that only when necessary */
  module->nb_bytes=(44100L*8*5/(module->bpm*2))&(~7); /* the & ~7 because a datum is 8 bytes long, so we want a 8 bytes
						       * alignment
						       */
  module->global_volume=64;

  module->current_line--;
  module->speed=32768;
  module->to_handle=new_line(module);
  module->new_line=module->new_pattern=0;
}

/* you give an instrument, a note, and it tells you what increment to take.
 * The first 32b are integere-increment, the next 32b are the after the dot
 * example : if increment=123.456, then the 32b first are 123 and the 32next are 2^32*.456
 */
long long get_increment(instrument *inst, int note)
{
  unsigned long long ret;
  long double period, freq, fl;
  sample *the_sample=&inst->sample_headers[(int)inst->sample_number[note]];
  char fine=the_sample->sample_fine_tune;//-128;
  note+=the_sample->sample_relative_note_number;
  period=10.*12*16*4- (long double)note*16*4 - (long double)fine/2.;
  freq=8363. * pow(2, (6*12*16*4 - period)/(12*16*4));
  //fprintf(stderr, "note = %d freq = 8363*2^(%Lf) = %Lf\t", note, (6*12*16*4 - period)/(12*16*4), freq);
  /* if 16b sample, then 2 bytes per datum, so a double freq */
  if (the_sample->sample_type & 0x10)
    freq*=2.;
  freq/=44100.;
  //fprintf(stderr, "freq = %Lf\t", freq);
  fl=floor(freq);
  //fprintf(stderr, "fl = %Lf\t", fl);
  freq-=fl;
  //fprintf(stderr, "freq = %Lf\t", freq);
  freq*=(unsigned long long)0xffffffff;
  freq=rint(freq);
  ret=(((unsigned long long)fl)<<32)|((unsigned long)freq);
  //fprintf(stderr, "ret normal = %lld ret haut = %lld ret bas = %lld\n", ret, ret>>32, ret & 0xffffffff);
  return ret;
}

/* this function will read a new line in the sample, change global-vars consequently */
/* return the number of bytes the tick is (so, it will be called not only when we change line
 * but when we change tick, it's possible for a line to be more than one tick (up to 32 i think)
 */
int new_line(XM *module)
{
  channel *channels=module->channels;
  unsigned int the_pattern;
  int i, nb_rows;
  note *cur_note;

  /* here, we maybe only have to go on for one tick */

  if (module->speed<module->current_tempo)
    /* yes, it only was one tick, so don't change the line, but handle effects, so a little goto (i love those gotos) */
    goto dont_change_line;

  /* here, we gotta change of line */
  module->speed=0;

  /* well, don't know how to handle it... by now, i only need to fill the 32 channels with :
   *   - is there a note or no ?
   *   - if yes, then fill the data in the channel :
   *          - current pointer of the sample
   *          - increment (64b, long long, hehe)
   *          - that's all folks, no loop, no panning, no volume. Just to see if it works so
   */

  module->current_line++;
  module->new_line=1;

  //fprintf(stderr, "pat=%d lin=%d\n", module->current_pattern, module->current_line);

  if (module->current_line >= module->patterns_nb_rows[(int)module->patterns_order[module->current_pattern]]) {
    module->current_pattern++;
    module->new_pattern=1;
    module->current_line=0;
    if (module->current_pattern>=module->song_length) {
      /* it's over, we stop, later, we'll loop, except if dumping */
      //fprintf(stderr, "mixer.new_line stop xm is over\n");
      //exit(1);
      if (module->do_we_dump)
	return 0;
      module->current_pattern=module->restart;
    }
  }

  /***************** here, i have to not forget to test things, it core-dumps ******************/
  /* done, it does not core anymore (but am i really sure of it ?) */
  the_pattern=module->patterns_order[module->current_pattern];
  nb_rows=module->patterns_nb_rows[the_pattern];
  if (!module->patterns) {
    fprintf(stderr, "error, pattern not existing\n");
    exit(1);
  }
  cur_note=module->patterns[the_pattern]+module->current_line;
  /* ok, let's run through the channels */
  /* we just gonna put the instrument to play, set current position to 0, that's all */
  for (i=0; i<module->nb_channels; i++, cur_note+=nb_rows) {
    /* if channel is closed, pass it over */
    if (channels[i].closed)
      continue;
    //fprintf(stderr, "n %d i %d\t", cur_note->note, cur_note->instrument);
    /* test cut-note, if envelop is there with sustain */
    if (cur_note->note==96)
      if (channels[i].instrument && (channels[i].instrument->volume_type & 1) && (channels[i].instrument->volume_type & 2))
	channels[i].cut_key=1;
      else
	channels[i].in_use=0;
    if (cur_note->note==0xff)
      goto ici_ici_ici;
    if (cur_note->instrument==-1)
      goto ici_ici_ici;
    /* we load a new note, set all parameters */
    if (!(channels[i].instrument=module->instruments[(int)cur_note->instrument])) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample=&channels[i].
      instrument->sample_headers[(int)channels[i].instrument->sample_number[(int)cur_note->note]];
    if (channels[i].sample->sample_data==(char *)0) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample_position=0;
    channels[i].increment=get_increment(module->instruments[(int)cur_note->instrument], cur_note->note);
    channels[i].in_use=1;
    channels[i].volume_point=0;
    channels[i].panning_point=0;
    channels[i].envelop_volume=channels[i].instrument->volume_type & 1 ? channels[i].instrument->volume_envelop[0][1] : 64;
    channels[i].envelop_panning=channels[i].instrument->panning_type & 1 ?
      channels[i].instrument->panning_envelop[0][1] : 32;
    channels[i].volume=channels[i].sample->sample_volume;
    channels[i].panning=channels[i].sample->sample_panning;
    channels[i].cut_key=0;
ici_ici_ici:
    if (cur_note->volume>=0x10 && cur_note->volume<=0x50)
      channels[i].volume=cur_note->volume-0x10;
  }
  //fprintf(stderr, "\n");
  //new_line();
  //exit(1);

  /* we've changed the line, let's handle the effects */

dont_change_line:
  /* let's handle the volume / panning */
  for (i=0; i<module->nb_channels; i++) {
    instrument *ins;

    if (!channels[i].in_use)
      continue;

    /* first, set the left/right volume according to the current volume / panning setting */
    ins=channels[i].instrument;

    {
      /* calculate final panning according to xm.txt */
      int pan=(unsigned char)channels[i].panning;
      int envpan=channels[i].envelop_panning;
      int final_panning=pan + ((envpan-32)*(128-abs(pan-128)))/32;

      channels[i].left_volume=(channels[i].volume*module->global_volume*
			       channels[i].envelop_volume*(255-final_panning))>>(6+6+8);
      channels[i].right_volume=(channels[i].volume*module->global_volume*
				channels[i].envelop_volume*final_panning)>>(6+6+8);
    }

    /* now, move the volume point */
    if (!(ins->volume_type & 1))
      goto deal_with_panning;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->volume_type & 2) &&
	(channels[i].volume_point == ins->volume_envelop[(int)ins->volume_sustain_point][0]))
      goto deal_with_panning;

    /* was it the loop point ? */
    if (ins->volume_type & 4) {
      if (channels[i].volume_point == ins->volume_envelop[(int)ins->volume_loop_end][0]) {
	channels[i].volume_point=ins->volume_envelop[(int)ins->volume_loop_start][0];
	channels[i].envelop_volume=ins->volume_envelop[(int)ins->volume_loop_start][1];
	goto deal_with_panning;
      }
    }

    /* or the end ? */
    if (channels[i].volume_point == ins->volume_envelop[ins->nb_volume_points-1][0])
      goto deal_with_panning;

    channels[i].volume_point++;
    /* set the envelop volume (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_volume_points; right++)
	if (ins->volume_envelop[right][0]>=channels[i].volume_point)
	  break;
      /* right is the first point on the right of the current volume point, we interpolate to find the new volume */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      x=channels[i].volume_point;
      x1=ins->volume_envelop[right-1][0];
      y1=ins->volume_envelop[right-1][1];
      x2=ins->volume_envelop[right][0];
      y2=ins->volume_envelop[right][1];
      x-=x1;
      x2-=x1;
      y2-=y1;
      channels[i].envelop_volume = y1 + x*y2/x2;
    }

deal_with_panning:
    /* now, move the panning point */
    if (!(ins->panning_type & 1))
      continue;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->panning_type & 2) &&
	(channels[i].panning_point == ins->panning_envelop[(int)ins->panning_sustain_point][0]))
      continue;

    /* was it the loop point ? */
    if (ins->panning_type & 4) {
      if (channels[i].panning_point == ins->panning_envelop[(int)ins->panning_loop_end][0]) {
	channels[i].panning_point=ins->panning_envelop[(int)ins->panning_loop_start][0];
	channels[i].envelop_panning=ins->panning_envelop[(int)ins->panning_loop_start][1];
	continue;
      }
    }

    /* or the end ? */
    if (channels[i].panning_point == ins->panning_envelop[ins->nb_panning_points-1][0])
      continue;

    channels[i].panning_point++;
    /* set the envelop panning (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_panning_points; right++)
	if (ins->panning_envelop[right][0]>=channels[i].panning_point)
	  break;
      /* right is the first point on the right of the current panning point, we interpolate to find the new panning */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      x=channels[i].panning_point;
      x1=ins->panning_envelop[right-1][0];
      y1=ins->panning_envelop[right-1][1];
      x2=ins->panning_envelop[right][0];
      y2=ins->panning_envelop[right][1];
      x-=x1;
      x2-=x1;
      y2-=y1;
      channels[i].envelop_panning = y1 + x*y2/x2;
      //      channels[i].panning <<=2;
    }
  }

  /* we've done a click, let's inc the speed (which is the number of ticks treated) */
  module->speed++;
  return (44100L*8*5/(module->bpm*2))&(~7); //module->nb_bytes;
}

/* this function will put in the soft mix-buffer exactly one datum, ie 8 bytes (32b mixer + stereo = 4*2 = 8 bytes) */
void handle_one_datum(XM *module)
{
  channel *channels=module->channels;
  int i;
  /* for all channels of the module, see if it's currently in use, then take the datum from the sample,
   * and add it to the mixer-buffer with correct left and right volume (and take care if 8 or 16b sample).
   * Then, increment the current pos in the sample, deal with loop or end of sample. */
  for (i=0; i<module->nb_channels; i++) {
    sample *sh;
    int pos;
    if (!channels[i].in_use)
      continue;
    sh=channels[i].sample;
    pos=channels[i].sample_position>>32;
    if (sh->sample_type & 0x10) { /* 16b sample */
      short *sample_data;
      sample_data=(short *)sh->sample_data;
      sample_data+=(pos>>1);
      mixer_soft_buffer[mixer_pos]+=((int)*sample_data*channels[i].left_volume)>>6;
      mixer_soft_buffer[mixer_pos+1]+=((int)*sample_data*channels[i].right_volume)>>6;
    } else { /* 8b sample */
      signed char *sample_data;
      sample_data=(signed char *)sh->sample_data;
      sample_data+=pos;
      /* it's -128 .. 127 and we want -32768 .. 32767, so <<8 */
      mixer_soft_buffer[mixer_pos]+=((((int)*sample_data)<<8)*channels[i].left_volume)>>6; 
      mixer_soft_buffer[mixer_pos+1]+=((((int)*sample_data)<<8)*channels[i].right_volume)>>6;
    }
    /* now, let's deal with loop or end */
    channels[i].sample_position+=channels[i].increment;
    pos=channels[i].sample_position>>32;
    if(!(sh->sample_type&3)) {
      /* no loop, so test end of sample, and if yes then channel no more in use */
      if (pos>=sh->sample_length) {
#if 0
	channels[i].in_use=0;
	/* we might have a problem with 2 cut-key played the one after the other. the next reset avoid it. */
	channels[i].instrument=(intrument *)0;
	/* no use, but i prefer to have (so if trouble is there, it will appear one day because of this null-pointer */
	channels[i].sample=(sample *)0;
#endif
	/* we clear all, that's simplier */
	memset(&channels[i], 0, sizeof(channel));
      }
    } else {
      /* let's test loop (ping-pong not yet supported) */
      if (pos>=sh->sample_loop_start+sh->sample_loop_length)
	channels[i].sample_position=(long long)sh->sample_loop_start<<32;
    }
  }
  mixer_pos+=2;
}

/* this function mix at most 1024 bytes or stops at the end of a tick */
int handle_tick(XM *module, int max_handle)
{
  int done=0;

  if (module->to_handle==0)
    module->to_handle=new_line(module);
  if (module->to_handle==0)
    /* if here, we are in dump-mode (we dump to wav), so we don't loop. We then return 0 */
    return 0;

  while(done<max_handle && module->to_handle>0) {
    handle_one_datum(module);
    /* because we have a 32b stereo mixer, so 8 bytes is one datum */
    done+=8;
    module->to_handle-=8;
  }
  return done;
}

void clear_mixer_soft_buffer(void)
{
  memset(mixer_soft_buffer, 0, sizeof(mixer_soft_buffer));
  mixer_pos=0;
}

void translate_mixed_data(void)
{
  int i;
  int *from=(int *)mixer_soft_buffer;
  short *to=(short *)mixer_buffer_soundcard;
  for (i=0; i<1024/2; i++, to++, from++)
    *to=(*from)>>3;
}

/* this function has to write 1024 bytes to the soundcard */
void mix(XM *module)
{
  int nb=1024*2, done;
  clear_mixer_soft_buffer();
  if (module->mix & MODULE_PLAY)
    while(nb) {
      done=handle_tick(module, nb);
      if (done==0) {
	/* if there, we stop the dump, and say how many bytes the last chunk is */
	module->do_we_dump=0;
	module->dumped=(1024*2-nb)>>2;
	return;
      }
      nb-=done;
    }
  /* we don't play anymore when dumping */
  if (module->do_we_dump) {
    module->dumped=512;
    return;
  }
  nb=1024*2;
  mixer_pos=0;
  if (module->mix & LIVE_PLAY)
    while(nb)
      nb-=handle_tick_live(module, nb);
  translate_mixed_data();
  play_sound(mixer_buffer_soundcard, 1024);
}

/* fun called to close all the channels, when we change the current playing module for example */
void close_all_channels(XM *xm)
{
  int i;

  for (i=0; i<32; i++)
    xm->channels[i].in_use=0;
}

/* fun used in sub_interface.interface_subwin_14, when we dec the nb of chan., we must close the "deleted" */
void close_two_channels(int nb, XM *xm)
{
  xm->channels[nb].in_use=xm->channels[nb+1].in_use=0;
}
